
/* 
   Configuration for compiling MzScheme

   If you want to set all the flags externally (on the command line
   with -D or some other compiler-dependent way), then define
   FLAGS_ALREADY_SET, and this file will be ignored.

   The best flag setings are already provided for some known
   architecture/systems. Otherwise, the default settings are generic
   Unix.  Send other architecture/system-specific info to 
   "mflatt@cs.rice.edu".
*/

#ifndef FLAGS_ALREADY_SET

/************** (BEGIN KNOWN ARCHITECTURE/SYSTEMS) ****************/

/************** Sun ****************/
#if defined(sun)

# define TIME_SYNTAX
# define PROCESS_FUNCTION
# define DIR_FUNCTION
# define GETENV_FUNCTION

# define HAS_STANDARD_IOB
# define FILES_HAVE_FDS

# define STACK_GROWS_DOWN
# define DO_STACK_CHECK
# define UNIX_FIND_STACK_BOUNDS

# define UNIX_DYNAMIC_LOAD

# define UNISTD_INCLUDE

# define FLAGS_ALREADY_SET

/************** x86 Linux ****************/
#elif defined(linux)

# define TIME_SYNTAX
# define PROCESS_FUNCTION
# define DIR_FUNCTION
# define GETENV_FUNCTION

# define HAS_LINUX_IOB
# define FILES_HAVE_FDS

# define STACK_GROWS_DOWN
# define DO_STACK_CHECK
# define UNIX_FIND_STACK_BOUNDS

# define FLAGS_ALREADY_SET

/************** Win32 under Borland C++ ****************/
#elif defined(__BORLANDC__)

# define TIME_SYNTAX
# define GETENV_FUNCTION
# define DIR_FUNCTION

# define STACK_GROWS_DOWN

# define DO_STACK_CHECK
# define USE_STACKAVAIL
# define STACK_SAFETY_MARGIN 25000

# define WIN32S_HACK
# define DIR_INCLUDE
# define USE_FTIME
# define IO_INCLUDE
# define RAND_NOT_RANDOM
# define NO_SLEEP

# define FLAGS_ALREADY_SET

#endif

/************** (END KNOWN ARCHITECTURE/SYSTEMS) ****************/

#ifndef FLAGS_ALREADY_SET

/*********************/
/* Language Features */
/*********************/

#define TIME_SYNTAX
#define PROCESS_FUNCTION
#define DIR_FUNCTION
#define GETENV_FUNCTION

 /* TIME_SYNTAX adds the (time ...) syntax; this may need to be
     turned off for compilation on some systems */

 /* PROCESS_FUNCTION add (process ...) and (system ...) functions */

 /* DIR_FUNCTION add (current-directory ...) function */

 /* GETENV_FUNCTION add (getenv ...) function */

/***********************/
/*       Ports         */
/***********************/

/* These are flags about the implementation of char-ready? for FILE*s
   None of these flags are required, but char-ready? may return
   spurious #ts if they are set up incorrectly. */

#define HAS_STANDARD_IOB
#define FILES_HAVE_FDS

 /* HAS_STANDARD_IOB, HAS_GNU_IOB, HAS_LINUX_IOB, and HAS_BSD_IOB
     are mutually exclusive; they describe how to read the FILE* 
     structure to determine if there are available cached characters. */

 /* FILES_HAVE_FDS means that a FILE* is always associated with a
    file desciptor, which can be select-ed to see if there are
    pending bytes. Don't use this unless one of the HAS_<X>_IOB
    flags is used. */

/***********************/
/* Stack Maniuplations */
/***********************/

# define DO_STACK_CHECK
# define UNIX_FIND_STACK_BOUNDS
# define STACK_SAFETY_MARGIN 50000

 /* STACK_GROWS_UP means that deeper stack values have higher
     numbered addresses.
    STACK_GROWS_DOWN means that deeper stack values have lower
     numbered addresses. This is usually the case (Sparc and
     Intel platforms, for example, use this).
    Use only one or none of these. (It's faster if you know which
     one applies, but it can also be figured it out dynamically.) */

 /* DO_STACK_CHECK checks for stack overflow during execution.
     Requires either UNIX_FIND_STACK_BOUNDS or USE_STACKAVAIL.  Not
     useful in most Unix systems (including Linux and SGI's Unix)
     where the stack is essentially unbounded. */

 /* UNIX_FIND_STACK_BOUNDS figures out the maximum stack position
     on Unix systems, using getrlimit() and the GC_find_stack_base()
     defined in the conservative garbage collector.
    USE_STACKAVIL uses stackavail() function for checking stack
     overflow; works with Borland C++, maybe other compilers.
    Use only one of these if DO_STACK_CHECK is used, or none otherwise. */

 /* STACK_SAFETY_MARGIN <X> sets the number of bytes that should be
     available on the stack for "safety" to <X>. Used only if
     DO_STACK_CHECK is used. STACK_SAFETY_MARGIN defaults to 50000. */

 /* ERROR_ON_OVERFLOW causes MzScheme to produce an error if the
     stack is overflowed. Normally, it would copy out the current
     stack and try to continue the computation. Used only if
     DO_STACK_CHECK is used. */

/***********************/
/*   Dynamic Loading   */
/***********************/

 /* UNIX_DYNAMIC_LOAD implements dynamic extensions under Unix
     using dlopen(); you may have to add the -ldl flag in the LIBS 
     Makefile variable. The library doesn't exist under Linux without 
     ELF, so it won't work. If you get linker errors about dl_open, etc., 
     this flag and the -ldl flag are the things to adjust.
    WINDOWS_DYNAMIC_LOAD implements dynamic extensions under Windows NT
     (Thanks to Patrick Barta)
    Use only one or none of these. */

/***********************/
/*    Miscellaneous    */
/***********************/

#define UNISTD_INCLUDE
#define RAND_NOT_RANDOM

 /* INEXACT_PRINT_DIGITS "<X>" uses <X> as the number of digits to
     use for printing floating-points. Defaults to "14" */

 /* RAND_NOT_RANDOM uses the function rand() instead of random()
     for random numbers. Some systems don't have random(). */

 /* DONT_IGNORE_FPE_SIGNAL if you don't want to ignore floating-point
     exception signals, or if your platform does not support signal() */

 /* NO_USER_BREAK_HANDLER turns off handling of INT signal in main.c;
     set this flag if there is no signal() function */

 /* DIR_INCLUDE if there's a <dir.h> file (mainly for MSWindows) */

 /* IO_INCLUDE if there's a <io.h> file (mainly for MSWindows) */

 /* UNISTD_INCLUDE if there's a <unistd.h> file (mainly for Unix) */

 /* USE_FTIME uses ftime instead of gettimeofday; only for TIME_SYNTAX */

 /* USE_ULIMIT uses ulimit instead of getdtablesize; only in main.c */

 /* NO_SLEEP means that there is no sleep() function */

 /* NO_STAT means that there is no stat() function */

 /* WIN32S_HACK uses a special hack to implement threads under Win32s. */

#endif  /* FLAGS_ALREADY_SET */
#endif  /* FLAGS_ALREADY_SET */
