/*
  MzScheme
  Copyright (c) 1995 Matthew Flatt
  All rights reserved.

  Please see the full copyright in the documentation.

  libscheme
  Copyright (c) 1994 Brent Benson
  All rights reserved.

  Thanks to Patrick Barta for the WINDOWS_DYNAMIC_LOAD code.
*/
#include "schpriv.h"
#include "schvers.h"
#include "../gc/gc.h"
#ifdef UNIX_DYNAMIC_LOAD
#include <dlfcn.h>
#elif defined(WINDOWS_DYNAMIC_LOAD)
#include <windows.h>
#endif

#include "schemex.h"

static Scheme_Object *load_extension(int argc, Scheme_Object **argv);

Scheme_Extension_Table *scheme_extension_table;

void scheme_init_dynamic_extension(Scheme_Env *env)
{
  if (scheme_starting_up) {
  scheme_extension_table=scheme_malloc_atomic(sizeof(Scheme_Extension_Table));
#include "schemex.inc"
  }

  scheme_add_global_constant("load-extension", 
			     scheme_make_prim_w_arity(load_extension, 1, 1), 
			     env);
}

void scheme_load_extension(const char *filename)
{
#ifdef UNIX_DYNAMIC_LOAD
  void *dl;
  char *(*f)(Scheme_Extension_Table *t), *vers;
  void (*f2)(Scheme_Env *);

  dl = dlopen(filename, 1);
  if (!dl)
    scheme_signal_error("load-extension: couldn't open \"%s\" (%s)",
			filename, dlerror());

  f = (char*(*)(Scheme_Extension_Table *))dlsym(dl, 
						"scheme_initialize_internal");

  if (!f)
    scheme_signal_error("load-extension: \"%s\" is not an extension (%s)", 
			filename, dlerror());

  vers = f(scheme_extension_table);
  if (!vers || strcmp(vers, VERSION))
    scheme_signal_error("load-extension: bad version %s from \"%s\"",
			vers, filename);

  f2 = (void (*)(Scheme_Env *))dlsym(dl, "scheme_initialize");

  if (!f2)
    scheme_signal_error("load-extension: no initializer in \"%s\" (%s)", 
			filename, dlerror());
  
  f2(scheme_env);
#elif defined(WINDOWS_DYNAMIC_LOAD)
  HINSTANCE dl;
  char *(*f)(Scheme_Extension_Table *t), *vers;
  void (*f2)(Scheme_Env *);
  
  dl = LoadLibrary(filename);
  if (!dl)
    scheme_signal_error("load-extension: couldn't open \"%s\"",
			filename);

  f = (char*(*)(Scheme_Extension_Table *))GetProcAddress(dl, 
						"scheme_initialize_internal");

  if (!f)
    scheme_signal_error("load-extension: couldn't open \"%s\"",
			filename);

  vers = (f)(scheme_extension_table);
  if (!vers || strcmp(vers, VERSION))
    scheme_signal_error("load-extension: bad version %s from \"%s\"",
			vers, filename);

  f2 = (void (*)(Scheme_Env *))GetProcAddress(dl,"scheme_initialize");

  if (f2 == NULL)
    scheme_signal_error("load-extension: no initializer in \"%s\"", 
			filename);
  
  (f2)(scheme_env);
#else
   scheme_signal_error("load-extension: not supported on this platform");
#endif
}


void scheme_register_extension_global(void *ptr, long size)
{
  GC_add_roots(ptr, (void *)(((char *)ptr) + size + 1));
}

static Scheme_Object *load_extension(int argc, Scheme_Object **argv)
{
  if (argc != 1)
    scheme_wrong_count("load-extension", 1, 1, argc, argv);
  if (!SCHEME_STRINGP(argv[0]))
    scheme_wrong_type("load-extension", "string", 0, argc, argv);
  
  scheme_load_extension(SCHEME_STR_VAL(argv[0]));

  return scheme_void;
}
